;-------------------------------------------------------------------------------
; SoftMPU 0.51 by bjt - Software MPU-401 emulator
;-------------------------------------------------------------------------------
;
; Uses the EMM386 I/O port trapping API to intercept reads and writes to a real
; MPU interface that only supports UART mode.
;
; Supported commands (reset, change to UART mode) are passed unchanged while
; other commands queue an ACK for the next time the interface is read. This
; subset of full MPU functionality is sufficient to get many DOS games to output
; MIDI.
;
; TODO: Emulation of MPU interrupts
; TODO: Intelligent-mode emulation (based on Dosbox code?)
; TODO: SYSEX throttling for Rev.0 MT-32
;
; CHANGELOG
;
; 0.5   Initial release
; 0.51  Fixed ID string search segment offsets
;       Fixed ACK behaviour after 'fake' UART switch
;
;-------------------------------------------------------------------------------

                .model  tiny
                .386

                text segment dword public use16
		assume cs:text,ds:text
		org 100h

Start:          jmp     Init

PortTable       label   word
DataPortAddr    dw      0                       ; Set from cmdline
DataHandlerAddr dw      offset PortHandler
CmdPortAddr     dw      0                       ; Set from cmdline
CmdHandlerAddr  dw      offset PortHandler

PortHandler:    ; Save registers
		push    dx

		; Test for input/output
		test    ecx,ecx
		jz      @@HandleInput

		; Handle output
                cmp     dx,DataPortAddr
		je      @@ExitHandler           ; Output on data port

		cmp     al,0FFh
		je      @@SetReset
		cmp     al,03Fh
		je      @@SetRealUART
		cmp     UARTMode,0              ; Already in UART mode?
		je      @@SetFakeUART
		mov     DoAnACK,1               ; Some other command, queue an ACK
		jmp     @@ExitHandlerVirt       ; Don't call original I/O handler

@@SetReset:     cmp     UARTMode,2              ; In fake UART mode?
		jne     @@ResetNoACK
		mov     DoAnACK,1               ; In fake UART mode so ACK
@@ResetNoACK:   mov     UARTMode,0              ; Hardware might ACK
		jmp     @@ExitHandler           ; Call original I/O handler

@@SetRealUART:  cmp     UARTMode,2              ; In fake UART mode?
		jne     @@UARTNoACK
		mov     DoAnACK,1               ; In fake UART mode so ACK
@@UARTNoACK:    mov     UARTMode,1              ; Hardware might ACK
		jmp     @@ExitHandler

@@SetFakeUART:  mov     UARTMode,2
		mov     al,03Fh                 ; Set UART command
		out     dx,al
		jmp     @@ExitHandlerVirt

@@HandleInput:  cmp     dx,DataPortAddr
		je      @@MidiDataIn            ; Read data port
                cmp     DoAnACK,1
		jne     @@ExitHandler           ; Call original I/O handler
		in      al,dx                   ; Read status port manually
		and     al,07Fh                 ; We have ACK waiting so clear DSR
		jmp     @@ExitHandlerVirt       ; Virtualised input

@@MidiDataIn:   cmp     DoAnACK,1
		jne     @@ExitHandler           ; Don't need to ACK so just exit
                mov     DoAnACK,0
		mov     al,0FEh
		jmp     @@ExitHandlerVirt       ; Virtualised input
	    
@@ExitHandler:  pop     dx
		stc                             ; Call original I/O handler
		retf

@@ExitHandlerVirt:
		pop     dx
		clc                             ; Virtualised, don't call original handler
		retf

DoAnACK         db      0
UARTMode        db      0                       ; 1 = Set by real UART command
						;     (pass next reset or UART to hardware)
						; 2 = Set by some other command
						;     (ACK on next reset or UART cmd)

		align   4
IDString        db      '0S2M'                  ; Detection string

Init:           ; Print banner
		mov     ah,9
                mov     dx,offset Header
		int     21h

		; Get command line
		mov     bx,080h
		xor     cx,cx
		mov     cl,[bx]
		test    cx,cx                   ; Any command line?
		jz      @@PrintHelp
		
		lea     si,[bx+1]
		cld                             ; Auto increment si
@@ParseCmdline: lodsd                           ; Get a dword from the cmd line
		cmp     eax,':TRO'              ; Check for base param
		je      @@BaseParam
                cmp     eax,':tro'
		jne     @@NotBaseParam
		
@@BaseParam:    cmp     cx,7                    ; Command plus 3-digit port
		jl      @@NotBaseParam          ; Not enough chars left
		xor     bx,bx
		mov     al,[si]                 ; 1st hex digit
		call    @@CharToNum
		or      bh,al
		mov     al,[si+1]               ; 2nd hex digit
		call    @@CharToNum
		or      bl,al
		shl     bl,4
		mov     al,[si+2]               ; Last hex digit
		call    @@CharToNum
		or      bl,al
                mov     DataPortAddr,bx         ; Set data port
		inc     bx                               
                mov     CmdPortAddr,bx          ; Set command port
		jmp     @@CheckPorts

@@NotBaseParam: sub     si,3                    ; Only step a byte at a time
		dec     cx                      ; We've consumed a char
		cmp     cx,0                    ; Consumed all chars?
		jg      @@ParseCmdline
		
@@CheckPorts:   ; Check we have a valid port range
		cmp     DataPortAddr,0200h
		jl      @@PrintHelp
		cmp     DataPortAddr,03FFh
		jg      @@PrintHelp

		; Check for EMM386
                push    ds
                push    0
                pop     ds
                mov     ax,ds:[019Ch]
                or      ax,ds:[019Eh]
                pop     ds
                jz      @@NoEmm
        
                mov     ax,0FFA5h
                int     67h
                cmp     ax,0845Ah
                jz      @@FoundEMM
		
@@NoEMM:        ; EMM386 not present
                mov     ah,9
                mov     dx,offset NoEMM
                int     21h
                jmp     @@Terminate

@@FoundEMM:     ; Check if already loaded
                mov     esi,offset IDString
                lodsd
                inc     eax                     ; Construct valid ID string
                mov     ebx,eax

		; Iterate through low 640k and look for string
		push    ds
		mov     esi,00000000h
@@CompareNext:  mov     ecx,esi
                and     ecx,0FFFF0000h
                shr     ecx,4
		mov     ds,cx                   ; Set segment base
		mov     eax,dword ptr ds:[si]   ; Use 16-bit offset only
		cmp     eax,ebx
		je      @@FoundID
		add     esi,4                   ; Increment dword offset
		cmp     esi,000A0000h           ; Stop at 640k
		jne     @@CompareNext
                pop     ds

                ; Set up port handler
                mov     ax,04A15h
		mov     bx,0
                mov     dx,CmdPortAddr
		shl     edx,16
                mov     dx,DataPortAddr         ; Cmd addr high, data addr low
		mov     cx,2                    ; Num ports
                mov     si,offset PortTable
                mov     di,offset Init          ; Offset to start
		int     2Fh
		jnc     @@EMMOK
		jmp     @@EMMErr
		
@@EMMOK:        mov     esi,offset IDString
		lodsd
                inc     eax
                mov     edi,offset IDString
		stosd                           ; Store complete ID string

		; Write the port address into our string
		mov     bx,DataPortAddr
		mov     ax,bx
		shr     ax,8
		call    @@NumToChar
		mov     LoadedPortNum,al
		mov     ax,bx
		shr     ax,4
		and     al,0Fh
		call    @@NumToChar
		mov     (LoadedPortNum+1),al
		mov     ax,bx
		and     al,0Fh
		call    @@NumToChar
		mov     (LoadedPortNum+2),al

                ; Print success message
                mov     ah,9
                mov     dx,offset LoadedStart
		int     21h
                mov     dx,offset Init
		int     27h                     ; Stay resident

@@PrintHelp:    mov     ah,9
		mov     dx,offset HelpText
		int     21h
		jmp     @@Terminate

@@FoundID:      pop     ds                      ; Restore segment base
		mov     ah,9
                mov     dx,offset LoadedError
		int     21h
		jmp     @@Terminate

@@EMMErr:       mov     ah,9
                mov     dx,offset EMMError
		int     21h
@@Terminate:    mov     ax,04C01h               ; Terminate program
		int     21h

@@CharToNum:    ; Input char in al
		cmp     al,061h
		jl      @@NotLower
		sub     al,057h                 ; Lowercase hex
		retn
@@NotLower:     cmp     al,041h
		jl      @@NotUpper
		sub     al,037h                 ; Uppercase hex
		retn
@@NotUpper:     sub     al,030h                 ; Digit
		retn

@@NumToChar:    ; Input num 0-F in al
		cmp     al,0Ah
		jl      @@NotLetter
		add     al,037h                 ; Uppercase hex
		retn
@@NotLetter:    add     al,030h                 ; Digit
		retn

Header          db      '---------------------------------------------------',0Dh,0Ah
                db      'SoftMPU 0.51 by bjt ',0FEh,' vogons.zetafleet.com for info',0Dh,0Ah
                db      '---------------------------------------------------',0Dh,0Ah
		db      0Dh,0Ah,'$'
NoEMM           db      'Error: EMM386 is not available',0Dh,0Ah,'$'
EMMError        db      'Error: Failed to set up port trap',0Dh,0Ah,'$'
LoadedError     db      'Error: SoftMPU is already loaded',0Dh,0Ah,'$'
LoadedStart     db      0Eh,' SoftMPU loaded at port '
LoadedPortNum   db      '   '
LoadedEnd       db      020h,0Eh,0Dh,0Ah,'$'
HelpText        db      'Usage: SOFTMPU /PORT:nnn',0Dh,0Ah
		db      0Dh,0Ah
		db      '    /PORT:nnn       Specifies the base I/O port address of the',0Dh,0Ah
		db      '                    MPU-401 compatible interface (200-3FF)',0Dh,0Ah,'$'

		text    Ends
		end     Start
